import threading
from wsgiref.simple_server import demo_app, make_server
from selenium import webdriver

from dojo import views


def before_all(context):
    context.server = make_server('', 8000, demo_app)
    context.server.set_app(views.app)
    context.thread = threading.Thread(target=context.server.serve_forever)
    context.thread.start()
    context.browser = webdriver.Firefox()


def after_all(context):
    context.server.shutdown()
    context.thread.join()
    context.browser.quit()



import unittest

from dojo.views import app


class FlaskrTestCase(unittest.TestCase):

    def setUp(self):
        app.config['TESTING'] = True
        self.app = app.test_client()

    def test_urls_are_available(self):
        URLS = ['/', '/task1', '/task2']
        for url in URLS:
            response = self.app.get(url)
            self.assertEqual(response.status_code, 200)
        response = self.app.post('/task1')
        self.assertEqual(response.status_code, 200)



# -*- coding: utf-8 -*-

from dojo.forms import SimpleForm
from dojo.task1 import Parser
from dojo.task2 import PlayGround
from flask import flash, Flask, render_template, request

app = Flask(__name__)
app.secret_key = 'really_very_secret'


NAVIGATION_BAR = [
    ('/', 'index', 'Index'),
    ('/task1', 'task1', 'Task1'),
    ('/task2', 'task2', 'Task2')
]


@app.route('/task2', methods=['GET', 'POST'])
def task2():
    app_name = "task2"
    description = "Bitte geben sie einen Text mit Spielfeldern ein"
    form = SimpleForm(request.form)
    context = {
        'active_page': app_name,
        'app_name': app_name,
        'description': description,
        'form': form,
        'navigation_bar': NAVIGATION_BAR,
    }
    if request.method == 'POST' and form.validate():
        playground = PlayGround(text=form.text.data)
        number_of_groups = [str(number_of_groups) for number_of_groups in playground.get_solution()]
        context['solution'] = "\n".join(number_of_groups)

    return render_template('task.html', **context)


@app.route('/task1', methods=['GET', 'POST'])
def task1():
    app_name = "task1"
    description = u"Bitte geben sie einen Text und die zu suchenden Wörter ein"
    form = SimpleForm(request.form)
    context = {
        'active_page': app_name,
        'app_name': app_name,
        'description': description,
        'form': form,
        'navigation_bar': NAVIGATION_BAR,
    }

    if request.method == 'POST' and form.validate():
        parser = Parser(text=form.text.data)
        parser.normalize()
        flash(u'Gesucht wird nach: {}'.format(parser.words))
        context['solution'] = parser.get_solution()

    return render_template('task.html', **context)


@app.route('/')
def index(app_name=None):
    return render_template('index.html', navigation_bar=NAVIGATION_BAR)


if __name__ == '__main__':
    app.run(debug=True)



# -*- coding: utf-8 -*-
from dojo.task1 import Parser

INPUT = u"""
Ein toller Beispieltext ist Blindtext. Er hat ein paar Wörter. Dies ist ein
Beispieltext, der ein paar Wörter hat und auch noch ein paar mehr, um die
Zeile etwas länger zu machen. Darüber hinaus ist er nur dafür da, um
genügend Testtext zusammenzubekommen. Dem Text selbst macht das nicht so
viel aus. Früher einmal mehr, als er noch nicht so selbstbewusst war. Heute
kennt er seine Rolle als Blindtext und fügt sich selbstbewusst ein. Er ist
ja irgendwie wichtig. Manchmal jedoch, ganz manchmal, weint er in der Nacht,
weil er niemals bis zum Ende gelesen wird. Doch das hat ja jetzt zum Glück
ein Ende.
5
ein
Beispieltext
der
paar
Wörter
"""

EXPECTED_TEXT = u"""
Ein toller Beispieltext ist Blindtext. Er hat ein paar Wörter. Dies ist ein
Beispieltext, der ein paar Wörter hat und auch noch ein paar mehr, um die
Zeile etwas länger zu machen. Darüber hinaus ist er nur dafür da, um
genügend Testtext zusammenzubekommen. Dem Text selbst macht das nicht so
viel aus. Früher einmal mehr, als er noch nicht so selbstbewusst war. Heute
kennt er seine Rolle als Blindtext und fügt sich selbstbewusst ein. Er ist
ja irgendwie wichtig. Manchmal jedoch, ganz manchmal, weint er in der Nacht,
weil er niemals bis zum Ende gelesen wird. Doch das hat ja jetzt zum Glück
ein Ende."""


def test_read_input():
    parser = Parser(text=INPUT)
    assert parser.text == EXPECTED_TEXT
    assert parser.words == [u'ein', u'beispieltext', u'der', u'paar', u'w\xf6rter']


def test_normalize_text():
    parser = Parser(text="12! 123.   3,o  5  ass?")
    parser.normalize()
    assert parser.text == ' o ass'


def test_normalize_words():
    parser = Parser()
    parser.words = ["12!", "123.", "3,o", "5  s?"]
    parser.normalize()
    assert parser.words == ['o', 's']


def test_get_rarest_word():
    parser = Parser()
    parser.words = ['aaa', 'bbb', 'ccc', 'ddd']
    parser.word_indicies = {
        'aaa': [1, 2, 3],
        'bbb': [4, 5],
        'ccc': [6],
        'ddd': [7],
    }
    assert parser.get_rarest_words() == ['ccc', 'ddd']


def test_get_solution():
    parser = Parser(text=INPUT)
    parser.normalize()
    solution = parser.get_solution()
    assert solution == "Beispieltext der ein paar Wrter"


EXAMPLE_INPUT = u"""
Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor
invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et
accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata
sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur
3
amet
sed
hollerididudeldoe
"""


def test_get_error_message():
    parser = Parser(text=EXAMPLE_INPUT)
    parser.normalize()
    solution = parser.get_solution()
    assert solution == "KEIN ABSCHNITT GEFUNDEN"


INPUT_IGNORE_CASE = u"""
Ein toller Beispieltext ist Blindtext. Er hat ein paar Wörter. Dies ist ein
Beispieltext, der ein paar Wörter hat und auch noch ein paar mehr, um die
Zeile etwas länger zu machen. Darüber hinaus ist er nur dafür da, um
genügend Testtext zusammenzubekommen. Dem Text selbst macht das nicht so
viel aus. Früher einmal mehr, als er noch nicht so selbstbewusst war. Heute
kennt er seine Rolle als Blindtext und fügt sich selbstbewusst ein. Er ist
ja irgendwie wichtig. Manchmal jedoch, ganz manchmal, weint er in der Nacht,
weil er niemals bis zum Ende gelesen wird. Doch das hat ja jetzt zum Glück
ein Ende.
5
Ein
BEISPIELTEXT
der
paaR
wörTer
"""


def test_get_solution_ignore_case():
    parser = Parser(text=INPUT_IGNORE_CASE)
    parser.normalize()
    solution = parser.get_solution()
    assert solution == "Beispieltext der ein paar Wrter"



# -*- coding: utf-8 -*-
from dojo.task2 import PlayGround

TEXT = """4
4
0010
1010
0100
1111
4
1001
0000
0110
1001
5
10011
00100
00000
11111
00000
8
00100100
10000001
00100101
01000100
10000000
00110110
10110110
00000000
"""


def test_playground_read_input():
    playground = PlayGround(text=TEXT)
    assert len(playground.fields) == 4


def test_playground_get_solution():
    playground = PlayGround(text=TEXT)
    assert playground.get_solution() == [1, 3, 3, 9]


def test_field_setup_cells():
    expected = [
        (0, 0), (1, 0), (2, 0), (3, 0),
        (0, 1), (1, 1), (2, 1), (3, 1),
        (0, 2), (1, 2), (2, 2), (3, 2),
        (0, 3), (1, 3), (2, 3), (3, 3),
    ]

    playground = PlayGround(text=TEXT)
    field = playground.fields[0]
    field.setup_cells()
    assert field.todo == expected


def test_field_get_neighbours_of_this_cell():
    playground = PlayGround(text=TEXT)
    field = playground.fields[0]

    expected = [(0, 1), (0, 2), (0, 3), (1, 1), (1, 3), (2, 1), (2, 2), (2, 3)]
    assert expected == sorted(field.get_neighbours_of_cell(1, 2))

    expected = [(0, 1), (1, 0), (1, 1)]
    assert expected == sorted(field.get_neighbours_of_cell(0, 0))

    expected = [(2, 0), (2, 1), (3, 1)]
    assert expected == sorted(field.get_neighbours_of_cell(3, 0))

    expected = [(0, 1), (0, 3), (1, 1), (1, 2), (1, 3)]
    assert expected == sorted(field.get_neighbours_of_cell(0, 2))

    expected = [(0, 2), (1, 2), (1, 3)]
    assert expected == sorted(field.get_neighbours_of_cell(0, 3))

    expected = [(2, 1), (2, 2), (2, 3), (3, 1), (3, 3)]
    assert expected == sorted(field.get_neighbours_of_cell(3, 2))

    expected = [(2, 2), (2, 3), (3, 2)]
    assert expected == sorted(field.get_neighbours_of_cell(3, 3))


def test_field_get_solution():
    playground = PlayGround(text=TEXT)
    field = playground.fields[0]
    assert field.get_solution() == 1



# -*- coding: utf-8 -*-
from behave import *
import time

VALID_INPUT = u"""
Ein toller Beispieltext ist Blindtext. Er hat ein paar Wörter. Dies ist ein
Beispieltext, der ein paar Wörter hat und auch noch ein paar mehr, um die
Zeile etwas länger zu machen. Darüber hinaus ist er nur dafür da, um
genügend Testtext zusammenzubekommen. Dem Text selbst macht das nicht so
viel aus. Früher einmal mehr, als er noch nicht so selbstbewusst war. Heute
kennt er seine Rolle als Blindtext und fügt sich selbstbewusst ein. Er ist
ja irgendwie wichtig. Manchmal jedoch, ganz manchmal, weint er in der Nacht,
weil er niemals bis zum Ende gelesen wird. Doch das hat ja jetzt zum Glück
ein Ende.
5
ein
Beispieltext
der
paar
Wörter
"""

INVALID_INPUT = u"""
schnick schnack schnick schnack schnick schnack schnick schnack schnick schnack
schnick schnackschnick schnackschnick schnackschnick schnack
1
schnuck
"""


@when('we visit the form and enter valid data')
def step1(context):
    context.browser.get('http://127.0.0.1:8000/task1')
    form = context.browser.find_element_by_id('text')
    form.send_keys(VALID_INPUT)
    form.submit()


@then('we see the solution')
def step2(context):
    solution = context.browser.find_element_by_css_selector('#solution pre')
    assert solution.text == "Beispieltext der ein paar Wrter"


@when('we visit the form and enter invalid data')
def step3(context):
    context.browser.get('http://127.0.0.1:8000/task1')
    form = context.browser.find_element_by_id('text')
    form.send_keys(INVALID_INPUT)
    form.submit()


@then('we see the error message')
def step4(context):
    solution = context.browser.find_element_by_css_selector('#solution pre')
    assert solution.text == "KEIN ABSCHNITT GEFUNDEN"



# -*- coding: utf-8 -*-
from behave import *

INPUT_NORMAL = u"""4
4
0010
1010
0100
1111
4
1001
0000
0110
1001
5
10011
00100
00000
11111
00000
8
00100100
10000001
00100101
01000100
10000000
00110110
10110110
00000000
"""

INPUT_EXTREME = u"""2
4
0000
0000
0000
0000
4
1111
1111
1111
1111
"""


@when('we visit the form of task2 and enter valid data')
def step1(context):
    context.browser.get('http://127.0.0.1:8000/task2')
    form = context.browser.find_element_by_id('text')
    form.send_keys(INPUT_NORMAL)
    form.submit()


@then('we see the solution for task2')
def step2(context):
    solution = context.browser.find_element_by_css_selector('#solution pre')
    assert solution.text == "1\n3\n3\n9"


@when('we visit the form of task2 and enter valid extreme data')
def step3(context):
    context.browser.get('http://127.0.0.1:8000/task2')
    form = context.browser.find_element_by_id('text')
    form.send_keys(INPUT_EXTREME)
    form.submit()


@then('we see the solution for extreme task2')
def step4(context):
    solution = context.browser.find_element_by_css_selector('#solution pre')
    assert solution.text == "0\n1"



# -*- coding: utf-8 -*-
import re
import logging
from collections import deque, defaultdict, namedtuple

Result = namedtuple('Result', 'distance begin end')
log = logging.getLogger('werkzeug')

is_one_digit = re.compile(r'\d$')
regex_text = re.compile(r'[^a-zA-Z\s]')
multiple_white_spaces = re.compile(r'\s+')


class Parser:
    """
    Class to find out the shortest earliest range in a given text, which contains the given words.

    """
    def __init__(self, text=''):
        self.word_indicies = defaultdict(list)
        self.words = []
        text_lines = []

        is_text = True
        for line in text.splitlines():

            if is_one_digit.match(line):
                is_text = False
                continue

            if is_text:
                text_lines.append(line)
            else:
                self.words.append(line.lower())

        self.text = "\n".join(text_lines)
        self.words = self.words

    def normalize(self):
        """
        Normalize a string means here: Remove everything but characters.
        Remove empty elements from the normalized result.

        """
        self.text = regex_text.sub('', self.text)
        self.text = multiple_white_spaces.sub(' ', self.text)

        self.words = [regex_text.sub('', word) for word in self.words]
        self.words = [multiple_white_spaces.sub('', word) for word in self.words]
        self.words = filter(None, self.words)

    def _get_positions_of_each_word_in_text(self, text):
        """
        Store the index of each word within the text in a dictionary. The key is the word.

        """
        indicies = defaultdict(list)
        for index, term in enumerate(text.split()):
            indicies[term.lower()].append(index)
        return indicies

    def get_solution(self):
        """
        Algorithm:
        - find all indicies for all words within the text
        - find out which words are the rarest
        - for each of the rarest words find out the smallest sum of the distances of each nearest
          occurence (absolute value) of each other word from the list of words to search for
        - collect for each rarest word the sum of all distances, the begin and the and of the range
          which contains all words
        - return the smallest distance with the smallest value for begin

        """
        self.word_indicies = self._get_positions_of_each_word_in_text(self.text)

        if not self.words:
            return u"ES KONNTEN KEINE WÖRTER IDENTIFIZIERT WERDEN"

        if not self.each_word_found():
            return u"KEIN ABSCHNITT GEFUNDEN"

        self.rarest_words = self.get_rarest_words()
        log.info("\nThe rarest words: {}".format(self.rarest_words))

        results = []
        for word in self.rarest_words:
            result = self.get_distance_begin_and_end(word)
            log.info("\nResult: {}\n".format(result))
            results.append(result)

        result = sorted(results)[0]
        log.info("Results: {}".format(sorted(results)))

        return " ".join(self.text.split()[result.begin:result.end])

    def get_distance_begin_and_end(self, word):
        """
        Find the nearest 'indicies' of the other 'search_words' based on the rarest search_word.
        Sum up the distance for each search_word over all search word indicies. Do this for every
        index of the rarest search word and return the begin and the end of the nearest indicies.

        """
        list_of_search_words = list(self.words)
        list_of_search_words.remove(word)
        result = {}

        result_distance = None

        for index in self.word_indicies[word]:
            distance = 0
            begin = 0
            end = 0

            log.info("\nDiscover '{}' at index: {}".format(word, index))
            for search_word in list_of_search_words:
                nearest_index = self.find_nearest_value_in_a_list(index,
                                                                  self.word_indicies[search_word])
                log.info("'{}'' nearest value to {} is {}".format(
                         search_word,
                         index,
                         nearest_index))

                if begin == 0:
                    begin = nearest_index

                begin = min(begin, nearest_index)
                end = max(end, nearest_index)
                distance += self.get_distance(index, nearest_index)

            log.info("distance: {} begin: {} end: {}".format(distance, begin, end))

            if result_distance is None or distance < result_distance:
                result = Result(distance, begin, end + 1)
                result_distance = distance

        return result

    def find_nearest_value_in_a_list(self, value, list_to_browse):
        """
        Return the value of a list which is the nearest to a given value.

        """
        index_of_the_nearest_value = min(range(len(list_to_browse)),
                                         key=lambda i: abs(list_to_browse[i]-value))
        return list_to_browse[index_of_the_nearest_value]

    def get_distance(self, value_a, value_b):
        """
        Return the distance between two given values.

        """
        return abs(int(value_a) - int(value_b))

    def get_rarest_words(self):
        """
        Return a list of words which have the shortest list of occurrence (inicies) within the text.

        """
        rarest_words = deque()
        occurrence = 0
        for word in self.words:
            if occurrence == 0:
                occurrence = len(self.word_indicies[word])
                rarest_words.append((word, occurrence))

            elif occurrence > len(self.word_indicies[word]):
                occurrence = len(self.word_indicies[word])
                rarest_words.clear()
                rarest_words.append((word, occurrence))

            elif occurrence == len(self.word_indicies[word]):
                rarest_words.append((word, occurrence))

            log.info("\"{}\" occures {} times in {}".format(
                     word,
                     len(self.word_indicies[word]),
                     self.word_indicies[word]))

        return [word[0] for word in rarest_words]

    def each_word_found(self):
        """
        Return true if each word of the words to search for was found in the text.

        """
        for word in self.words:
            if not word in self.word_indicies:
                return False

        return True



from setuptools import setup, find_packages

version = '0.5'

setup(name='thoreg_dojo',
      version=version,
      description="The coding dojo of Thoreg",
      long_description="""\
Some programming examples and reminder - coding for fun and profit""",
      classifiers=['Programming Language :: Python',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Topic :: Software Development :: User Interfaces'],
      keywords='dojo, thoreg, examples, flask, behave, tests',
      author='Thomas Rega',
      author_email='thoreg@gmail.com',
      url='https://github.com/thoreg/dojo',
      license='BSD',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=['behave',
                        'coveralls',
                        'Flask',
                        'Flask-WTF',
                        'pytest',
                        'pytest-cov',
                        'watchdog'],
      )



# -*- coding: utf-8 -*-
import logging
import re

log = logging.getLogger('werkzeug')

IS_ONE_DIGIT = re.compile(r'\d$')


class Field:
    """
    Square field of '0's and '1's.

    """
    def __init__(self, dimension, field_lines):
        self.dimension = int(dimension)
        self.content = field_lines
        self.todo = []

    def get_solution(self):
        self.setup_cells()
        return self.get_number_of_groups()

    def get_number_of_groups(self):
        """
        Walk through the whole field and count the groups of '1' which are connected.
        Connected means here these 'cells' are neighbours via N/NO/O/OS/S/SW/W/NW.

        """
        log.info("self.content: {}".format(self.content))
        log.info("todo: ".format(self.todo))

        number_of_groups = 0
        seen = set()
        to_check = set()

        while self.todo:
            x, y = self.todo.pop()
            if (x, y) in seen:
                continue

            if self.content[y][x] == '0':
                seen.add((x, y))
                continue

            log.info("\nNew Cell to check found: ({}, {})".format(x, y))
            to_check.add((x, y))
            number_of_groups += 1

            while to_check:
                (x, y) = to_check.pop()
                if (x, y) in seen:
                    log.info("   cell {} already seen".format((x, y)))
                    continue

                neighbours = self.get_neighbours_of_cell(x, y)
                log.info("   neighbours: {}".format(neighbours))
                for nx, ny in neighbours:
                    if self.content[ny][nx] == '1':
                        to_check.add((nx, ny))

                seen.add((x, y))

        log.info("Number_of_groups: {}".format(number_of_groups))
        return number_of_groups

    def setup_cells(self):
        """
        Build a matrix with all coordinates for all cells of the field.

        """
        self.todo = []
        for y in range(self.dimension):
            for x in range(self.dimension):
                self.todo.append((x, y))

    def get_neighbours_of_cell(self, x, y):
        """
        Return a list of tuples with the coordinates of all neighbours of the cell at (x, y).

        ---------------
        | NW | N | NO |
        ---------------
        |  W | x |  O |
        ---------------
        | SW | S | SO |
        ---------------

        """
        neighbours = set()

        minimal_x = minimal_y = 0
        maximal_x = maximal_y = self.dimension - 1

        # N
        neighbours.add((x, max(y - 1, minimal_y)))
        # NO
        neighbours.add((min(x + 1, maximal_x), max(y - 1, minimal_y)))
        # O
        neighbours.add((min(x + 1, maximal_x), y))
        # SO
        neighbours.add((min(x + 1, maximal_x), min(y + 1, maximal_y)))
        # S
        neighbours.add((x, min(y + 1, maximal_y)))
        # SW
        neighbours.add((max(x - 1, minimal_x), min(y + 1, maximal_y)))
        # W
        neighbours.add((max(x - 1, minimal_x), y))
        # NW
        neighbours.add((max(x - 1, minimal_x), max(y - 1, minimal_y)))

        # A cell can not be the neighbour of itself
        if (x, y) in neighbours:
            neighbours.remove((x, y))

        return sorted(neighbours)


class PlayGround:
    """
    A playground consists of sevral Field objects.

    """
    def __init__(self, text=''):
        self.fields = []
        for line_number, line in enumerate(text.splitlines()):
            if line_number == 0:
                continue

            # A new field begins
            if IS_ONE_DIGIT.match(line):
                dimension = int(line)
                field_lines = []
                lines_left = dimension
                continue

            if lines_left:
                field_lines.append(line)
                lines_left -= 1
                if lines_left == 0:
                    self.fields.append(Field(dimension, field_lines))

    def get_solution(self):
        solutions = []
        for field in self.fields:
            solutions.append(field.get_solution())

        return solutions
